settings = {
    # delay between characters appearing on screen (in milliseconds)
    'delay':130,

    # range of random time in milliseconds to change the delay; makes typing
    # look more natural
    'delay_variance':30,

    # movie mode: instead of waiting for key presses, play like a movie, -1
    # disables, otherwise value in milliseconds for delay between played steps
    # (like 'delay' field
    'movie_mode':-1,

    # xterm colour mode, anything but 256 gives default
    'colour':256,
}



# =============================================================================
# Actions
# =============================================================================

import random
from enum import Enum

from purdy.content import Animate

# =============================================================================

class AppendAll:
    """Appends the code to the code box. All work is done in the setup()
    phase, does not wait for a key press.
    """
    def __init__(self, code_box, code_blob):
        """Constructor

        :param code_box: a :class:`CodeBox` widget where the code will be 
            presented
        :param code: a :class:`Code` object containing code to post
        """
        self.code_box = code_box
        self.code_blob = code_blob

    def setup(self, settings):
        for line in self.code_blob.lines:
            self.code_box.add_line(line.markup)

    def next(self, key):
        raise StopIteration


class ReplaceAll(AppendAll):
    """Same as :class:`AppendAll` except it clears the code box before
    inserting the code.
    """
    def setup(self, settings):
        # Replace is just like Append, but clear the box first
        self.code_box.clear()
        return super().setup(settings)

# -----------------------------------------------------------------------------

class State(Enum):
    WAITING = 0
    TYPING  = 1


class AppendTypewriter:
    """Appends the code to the code box using a typewriter animation. Meant to
    be used with console-style (REPL) scripts. Prompts are printed immediately,
    anything after a prompt waits for a key press. Each key press will cause
    the typewriter animation to print out as if you are typing a command after
    the prompt. Results from a "command" appear immediately. 
    """
    def __init__(self, code_box, code_blob):
        """Constructor

        :param code_box: a :class:`CodeBox` widget where the code will be 
                         presented
        :param code: a :class:`Code` object containing code to post
        """
        self.code_box = code_box
        self.code_blob = code_blob

    def setup(self, settings):
        # delay settings
        self.typing_delay = settings['delay'] / 1000
        self.variance = settings['delay_variance']

        # get the chunks of typewriter animation for the first line
        self.chunks = self.code_blob.typewriter_chunks()

        # insert the first chunk from the first line in our CodeBox
        self.current_chunk = 0

        self.state = State.TYPING
        self.code_box.add_line()

    @property
    def delay_until_next_letter(self):
        vary_by = random.randint(0, 2 * self.variance) - self.variance
        return self.typing_delay + (vary_by / 1000)

    def next(self, key):
        if self.current_chunk >= len(self.chunks):
            raise StopIteration

        if self.state == State.TYPING and key is not None:
            # called due to a keypress, we're busy typing so ignore it
            return -1

        if self.state == State.WAITING:
            # got a keypress: start typing
            self.state = State.TYPING

        chunk = self.chunks[self.current_chunk]
        while True:
            if chunk == Animate.NEWLINE:
                self.code_box.add_line()
                self.current_chunk += 1
            elif chunk == Animate.DELAY:
                self.current_chunk += 1
                return self.delay_until_next_letter
            elif chunk == Animate.WAIT:
                self.state = State.WAITING
                self.current_chunk += 1
                return -1
            else:
                # display the markup in the chunk
                self.code_box.set_last_line(chunk)
                self.current_chunk += 1

            if self.current_chunk >= len(self.chunks):
                raise StopIteration

            chunk = self.chunks[self.current_chunk]


class ReplaceTypewriter(AppendTypewriter):
    """Same as :class:`AppendTypewriter` but clears the code box before
    starting the animation.
    """
    def setup(self, settings):
        # Replace is just like Append, but clear the box first
        self.code_box.clear()
        super(ReplaceTypewriter, self).setup(settings)

# =============================================================================

class Insert:
    """Inserts a piece of code at the given line number. Code on the line
    number before the insertion is pushed downards -- i.e. inserting on line 1
    has the new code at the top of the document.
    """
    def __init__(self, code_box, line_number, code_blob):
        """Constructor

        :param code_box: a :class:`CodeBox` widget where the code will be 
            presented
        :param line_number: line number (1-indexed) where code is to be
                            inserted
        :param code: a :class:`Code` object containing code to post
        """
        self.code_box = code_box
        self.code_blob = code_blob
        self.line_number = line_number

    def setup(self, settings):
        line_no = self.line_number
        for line in self.code_blob.lines:
            self.code_box.add_line_at(line_no, line.markup)
            line_no += 1

        self.code_box.fix_line_numbers(self.line_number)

    def next(self, key):
        raise StopIteration

# =============================================================================

class Highlight:
    """Changes the colour attributes of the given line number to be
    highlighted.
    """
    def __init__(self, code_box, line_number, highlight):
        """Constructor

        :param code_box: a :class:`CodeBox` widget where the code will be 
            presented
        :param line_number: line number (1-indexed) where code is to be
                            inserted. Also supports a list of line numbers.
        :param highlight: `True` to turn highlighting on, `False` to turn it off
        """
        self.code_box = code_box
        self.line_number = line_number
        self.highlight = highlight

    def setup(self, settings):
        pass

    def next(self, key):
        numbers = self.line_number
        if isinstance(self.line_number, int):
            numbers = [self.line_number]

        for number in numbers:
            self.code_box.body.contents[number - 1].set_highlight(
                self.highlight)

        raise StopIteration

# =============================================================================

class StopMovie:
    """If the system is working in movie-mode, this action causes it to leave
    movie mode and go back to being interactive, expecting key presses.
    """
    def __init__(self, screen):
        """Constructor

        :param screen: :class:`Screen` object that controls the playback
        """
        self.screen = screen

    def setup(self, settings):
        pass

    def next(self, key):
        self.screen.movie_mode = -1

        raise StopIteration


class Wait:
    """Action does nothing but wait for a key press"""
    def __init__(self):
        self.called = False

    def setup(self, settings):
        pass

    def next(self, key):
        if not self.called:
            self.called = True
            return -1

        raise StopIteration



### content.py
#
# Methods for parsing and managing code to be displayed

import os
from collections import namedtuple
from copy import copy
from enum import Enum

from pygments.lexers import PythonConsoleLexer, PythonLexer
from pygments.token import Keyword, Name, Comment, String, Error, \
    Number, Operator, Generic, Token, Whitespace

# =============================================================================

class LexerContainer:
    def __init__(self):
        self._lexers = {
            'con':(PythonConsoleLexer(), 'Python Console'), 
            'py3':(PythonLexer(), 'Python 3'),
        }

    @property
    def default_name(self):
        return 'con'

    @property
    def choices(self):
        return self._lexers.keys()

    def get_lexer(self, name):
        return self._lexers[name][0]

LEXERS = LexerContainer()
    
# -----------------------------------------------------------------------------

class TokenLookup():
    colours = {
        # urwid colour spec supports both 16 and 256 colour terminals
        #                    fg16            bg16    fg256   bg256
        Token:              ('',             '', '', '',             ''),
        Whitespace:         ('',             '', '', '',             ''),
        Comment:            ('dark cyan',    '', '', 'dark cyan',    ''),
        Comment.Preproc:    ('dark cyan',    '', '', 'dark cyan',    ''),
        Keyword:            ('brown',        '', '', 'brown',        ''),
        Keyword.Type:       ('brown',        '', '', 'brown',        ''),
        Operator:           ('brown',        '', '', 'brown',        ''),
        Operator.Word:      ('brown',        '', '', 'brown',        ''),
        Name:               ('light gray',   '', '', 'light gray',   ''),
        Name.Builtin:       ('dark cyan',    '', '', '#068',         ''),
        Name.Function:      ('dark cyan',    '', '', 'light gray',   ''),
        Name.Namespace:     ('dark cyan',    '', '', 'light gray',   ''),
        Name.Class:         ('dark cyan',    '', '', 'light gray',   ''),
        Name.Exception:     ('dark green',   '', '', 'dark green',   ''),
        Name.Decorator:     ('dark cyan',    '', '', '#66d,bold',    ''),
        Name.Variable:      ('',             '', '', '',             ''),
        Name.Constant:      ('',             '', '', '',             ''),
        Name.Attribute:     ('',             '', '', '',             ''),
        Name.Tag:           ('',             '', '', '',             ''),
        String:             ('dark magenta', '', '', 'dark magenta', ''),
        Number:             ('dark magenta', '', '', 'dark magenta', ''),
        Generic.Prompt:     ('dark blue',    '', '', 'dark blue',    ''),
        Generic.Error:      ('dark green',   '', '', 'dark green',   ''),
        Generic.Traceback:  ('',             '', '', '#a00,bold',    ''),
        Error:              ('dark green',   '', '', '#f00',         ''),
    }

    @classmethod
    def get_colouring(cls, token):
        # Tokens are hierarchical and mapped to colours in our palette using
        # their names. If the token is in our colour map, return its name, if
        # it isn't, go up its hierarchy until a match is found
        if token in cls.colours:
            return str(token)

        # token not in our map, search its ancestors
        token = token.parent
        while(token != None):
            if token in cls.colours:
                return str(token)

            token = token.parent

        # something went wrong with our lookup, return the default
        return 'Token'

    @classmethod
    def is_a(cls, token1, token2):
        ### Returns true if token1 is equal to or a child of token2
        if token1 == token2:
            return True

        parent = token1.parent
        while(parent != None):
            if parent == token2:
                return True

            parent = parent.parent

        return False

    @classmethod
    def get_colour_attribute(cls, token):
        # colour attribute is a 6-tuple, first part being a name (we'll use
        # the same name as the token), next 5 parts are the colour strings for
        # fg16, bg16, x, fg256, bg256
        return ( str(token), ) + tuple( cls.colours[token] )

    @classmethod
    def get_highlight_colour_attribute(cls, token):
        # return 6-tuple colour attribute using the highlight colour as the
        # background colour instead of the one in our table
        return ( str(token) + '_highlight', 'black', 'light gray', '', 
            'white', 'g23' )

    @classmethod
    def line_number_markup(cls, line_number):
        return ('line_number', f'{line_number:3} ')

# =============================================================================

CodeToken = namedtuple('CodeToken', ['token_type', 'colour', 'text'])

class Animate(Enum):
    WAIT = 0
    DELAY = 1
    NEWLINE = 2


class CodeLine:
    """Represents a single line of code to be displayed. Wraps the urwid
    text markup language. The markup for this line can be accessed through 
    :attr:`CodeLine.markup`
    """
    def __init__(self, tokens):
        self.tokens = tokens
        self.markup = ['']
        self.is_output = self.tokens[0] == Generic.Output

        for token in self.tokens:
            if token.text == '':
                # tokenizer sometimes puts in empty stuff, skip it
                continue
            else:
                # append markup tuple to end of last row
                markup_tuple = (token.colour, token.text)
                if self.markup == ['']:
                    # was empty line, replace it with new markup
                    self.markup = [markup_tuple, ]
                else:
                    self.markup.append(markup_tuple)


class Code:
    """Object encapsulates text containing code that is to be displayed

    :param contents: string with lines of code
    :param lexer_name: name of lexer to use on the code, one of 'con' for
                       Console lexer or 'py3' for Python 3 lexer
    :param show_line_numbers: True if code should have line numbers when 
                              displayed, defaults to False
    """
    def __init__(self, contents, lexer_name, show_line_numbers=False):
        self.show_line_numbers = show_line_numbers
        lexer = LEXERS.get_lexer(lexer_name)

        # parse the code
        self.tokens = []
        self.lines = []
        for token_type, text in lexer.get_tokens(contents):
            colour = TokenLookup.get_colouring(token_type)
            self.tokens.append( CodeToken(token_type, colour, text) )

        CodeLineBuilder().parse(self.tokens, self.lines)

    def typewriter_chunks(self):
        return TypewriterChunkifier().parse(self.tokens)

# -----------------------------------------------------------------------------

class CodeFile(Code):
    """Subclass of :class:`Code` which reads code to be displayed from a file

    :param filename: name of file to read to get the code contents
    :param lexer_name: name of lexer to use on the code, one of 'con' for
                       Console lexer or 'py3' for Python 3 lexer
    :param show_line_numbers: True if code should have line numbers when 
                              displayed, defaults to False
    """
    def __init__(self, filename, lexer_name, show_line_numbers=False):
        filename = os.path.abspath(filename)
        with open(filename) as f:
            contents = f.read()

        super().__init__(contents, lexer_name, show_line_numbers)

# -----------------------------------------------------------------------------

class CodeLineBuilder:
    ### Class encapsulates handling tokens of code and turning them into
    # CodeLine objects with appropriate markup

    def _newline_handler(self, lines, token):
        # hit a CR, time to create a new CodeLine object
        if not self.token_set:
            token2 = CodeToken(token.token_type, 'empty', '')
            self.token_set = [token2 ]

        line = CodeLine(self.token_set)
        lines.append(line)

        # reset to start the next group of tokens
        self.token_set = []

    def _string_handler(self, lines, token):
        # String tokens may be multi-line
        for row in token.text.splitlines(True):
            token2 = CodeToken(token.token_type, token.colour, row.rstrip('\n'))
            self.token_set.append(token2)

            if row[-1] == '\n':
                line = CodeLine( self.token_set )
                lines.append(line)

                self.token_set = []

    def _default_handler(self, lines, token):
        if token.text[-1] == '\n':
            # there is a \n at the end of the token, need to rebuild it
            # without it, then create the CodeLine object
            token2 = CodeToken(token.token_type, token.colour,
                token.text.rstrip('\n'))
            self.token_set.append(token2)

            # token text caused a CR, create a new CodeLine object
            line = CodeLine(self.token_set)
            lines.append(line)

            # reset to start the next group of tokens
            self.token_set = []
        else:
            self.token_set.append(token)

    def parse(self, tokens, lines):
        self.token_set = []
        for token in tokens:
            if token.text == '\n':
                self._newline_handler(lines, token)
            elif token.text == '':
                # tokenizer sometimes puts in empty stuff, skip it
                continue
            elif TokenLookup.is_a(token.token_type, String) and \
                    '\n' in token.text:
                self._string_handler(lines, token)
            else:
                self._default_handler(lines, token)

# -----------------------------------------------------------------------------

class TypewriterChunkifier:
    ### Encapsulates token parsing into typewriter animation chunks

    def __init__(self):
        self.chunks = []
        self.line = []
        self.traceback = False

        self.handlers = {
            Generic.Prompt   :self._prompt_handler,
            Generic.Output   :self._output_handler,
            Generic.Traceback:self._traceback_handler,
        }

    def _prompt_handler(self, token):
        ### chunkifies Generic.Prompt tokens
        if self.traceback:
            # we're in traceback mode, spit out the last line then get
            # out of traceback mode
            self.traceback = False
            if self.line:
                self.chunks.append( copy(self.line) )
                self.line = []
        
        markup = (token.colour, token.text)
        self.line.append(markup)
        self.chunks.append( copy(self.line) )
        self.chunks.append( Animate.WAIT )

    def _output_handler(self, token):
        ### chunkifies Generic.Output tokens
        for row in token.text.rstrip('\n').split('\n'):
            self.chunks.append( (token.colour, row) )
            self.chunks.append( Animate.NEWLINE )

    def _traceback_handler(self, token):
        ### chunkifies Generic.Traceback tokens

        # tracebacks are funny, they are followed by stack trace information
        # that should be colourized but treated as a single output block, go
        # into "traceback mode" so that no animation happens until we get to
        # the next Prompt
        self.traceback = True
        text = token.text.rstrip()
        if text:
            self.chunks.append( (token.colour, token.text.rstrip() ) )
            self.chunks.append( Animate.NEWLINE )

        self.line = []

    def _default_handler(self, token):
        ### chunkifies tokens without specific handlers
        if self.traceback:
            self.line.append( (token.colour, token.text) )
        else:
            # type things letter by letter now
            word = ''
            for letter in token.text:
                word += letter
                markup = (token.colour, word)

                if len(word) == 1:
                    self.line.append( markup )
                else:
                    self.line[-1] = markup

                self.chunks.append( copy(self.line) )
                self.chunks.append( Animate.DELAY )

    def parse(self, tokens):
        for token in tokens:
            if token.text == '' and token.token_type != Generic.Traceback:
                # skip empty tokens
                continue

            if token.text == '\n':
                if self.line:
                    self.chunks.append( copy(self.line) )
                    self.line = []

                self.chunks.append( Animate.NEWLINE )
                continue

            # lookup the handler for this token and run it
            fn = self.handlers.get(token.token_type, self._default_handler)
            fn(token)

        # clean off any instructions from the end (e.g. if it ends in a Prompt
        # there will be a WAIT and a NEWLINE
        while isinstance(self.chunks[-1], Animate):
            self.chunks.pop()

        return self.chunks



import urwid

from purdy.content import TokenLookup
from purdy.settings import settings as default_settings

# =============================================================================
# Globals

palette = []
highlight_mapper = {}

# =============================================================================
# Main Screen
# =============================================================================

class BaseWindow(urwid.Pile):
    def __init__(self, screen, *args, **kwargs):
        self.screen = screen
        super(BaseWindow, self).__init__(*args, **kwargs)

    def _next_focus(self):
        # switch focus to the next focusable item in our pile

        # want to walk the list of widgets in the pile, starting at the one
        # after the one in focus, looping to the beginning
        #
        # build the list from the focus point to the end, then remove the
        # focus point
        indices = list(range(self.focus_position, len(self.contents)))
        indices.pop(0)

        # add from the beginning of the list to the focus point, inclusive
        indices.extend( range(0, self.focus_position + 1) )

        # now use the list of indicies to find the first widget that is
        # allowed to be focused and focus it
        for index in indices:
            widget, _ = self.contents[index]
            if getattr(widget, 'tab_focusable', False):
                # found a focusable widget, set it as focused and we're done
                self.focus_position = index
                return

    def keypress(self, size, key):
        if key in ('q', 'Q'):
            raise urwid.ExitMainLoop()

        if key not in ('right', 'left'):
            # we don't want to pass left/right to the children, we want it to
            # act as "next action", everything else pass to the child to
            # handle for scrolling etc
            result = super(BaseWindow, self).keypress(size, key)
            if result is None:
                # keypress was handled by child, we're done
                return None

        # --- at this point the keypress was not handled by the children, see
        # if we want to do anything with it
        if key == 'tab':
            self._next_focus()
            return None

        if not self.screen.actions:
            # no actions left to do, ignore the keypress
            return None

        if self.screen.movie_mode != -1:
            # in movie mode, ignore key press
            return None

        self.next_action(key)

    def alarm(self, loop, data):
        self.next_action()

    def next_action(self, key=None):
        """Actions are triggered either by a key press or an alarm. This
        method looks in the action queue and gets the tells the top most
        action to take its next step.

        :param key: the key that was pressed, or None if it was an alarm
        """
        try:
            # tell the current action to do its thing
            timer = self.screen.actions[0].next(key)

            if timer != -1:
                # positive value means set callback timer 
                self.screen.loop.set_alarm_in(timer, self.alarm)
            elif self.screen.movie_mode != -1:
                # in movie mode we simulate the key presses, set the callback
                self.screen.loop.set_alarm_in(self.screen.movie_mode, 
                    self.alarm)

        except StopIteration:
            # action is done doing things, pop it off the queue
            self.screen.actions.pop(0)

            if self.screen.actions:
                # there is another action in the action queue, set it up and
                # then force a call back to its first step
                self.screen.actions[0].setup(self.screen.settings)

                if self.screen.movie_mode != -1:
                    # in movie mode we simulate the key presses, set the 
                    # callback
                    self.screen.loop.set_alarm_in(self.screen.movie_mode, 
                        self.alarm)
                else:
                    self.screen.loop.set_alarm_in(0, self.alarm)


class Screen:
    """Manages the display and event loop for the slide show. All purdy
    scripts need to create one of these or its children.

    This class includes a single :class:`ui.CodeBox` widget which can be accessed
    as :attr:`Screen.code_box`
    """
    def __init__(self, conf_settings=None, show_line_numbers=False):
        """Constructor

        :param conf_settings: a settings dictionary object. Defaults to 
                              `None` which uses the default settings
                              dictionary: :attr:`settings.settings`

        :param show_line_numbers: True turns line numbers on inside the
                                  associated code box. Defaults to False
        """
        self.show_line_numbers = show_line_numbers
        self.settings = conf_settings
        if conf_settings is None:
            self.settings = default_settings

        self.movie_mode = self.settings['movie_mode']
        if self.movie_mode != -1:
            self.movie_mode = float(self.movie_mode) / 1000.0

        self._build_boxes()
        self._build_palette()

        self.loop = urwid.MainLoop(self.base_window, palette)

        if self.settings['colour'] == 256:
            self.loop.screen.set_terminal_properties(colors=256)
            self.loop.screen.reset_default_terminal_palette()

    def _build_boxes(self):
        self.code_box = CodeBox(self, self.show_line_numbers)
        self.base_window = BaseWindow(self, [self.code_box, ])

    def _build_palette(self):
        global palette, highlight_mapper

        for token in TokenLookup.colours.keys():
            # add the colour to the palette
            palette.append( TokenLookup.get_colour_attribute(token) )

            # add the highlight version of the colour to the palette
            palette.append( TokenLookup.get_highlight_colour_attribute(token) )

            # add a mapping between the colour and the highlight
            highlight_mapper[str(token)] = str(token) + '_highlight'

        # now append colour attributes that aren't for the tokens
        palette.extend([
            ('line_number', 'dark gray', '', '', 'dark gray', ''),
            ('empty', '', '', '', '', ''),
        ])

        # add default map value
        highlight_mapper[None] = 'highlight'

    def run(self, actions):
        """Calls the main display event loop. Does not return until the UI
        exits."""
        # store our display actions and setup the first one
        self.actions = actions
        self.actions[0].setup(self.settings)

        # as soon as the loop is going invoke the next action, setup a
        # callback
        if self.movie_mode != -1:
            # in movie mode we simulate the key presses, set the callback to
            # start the process
            self.loop.set_alarm_in(self.movie_mode, self.base_window.alarm)
        else:
            self.loop.set_alarm_in(0, self.base_window.alarm)

        # call urwid's main loop, this code doesn't return until the loop
        # exits!!!
        self.loop.run()


class SplitScreen(Screen):
    """Inheritor of :class:`Screen`. This implementation supports two 
    :class:`CodeBox` instances, stacked vertically and separated by a dividing
    line. The code boxes are :attr:`SplitScreen.top_box` and
    :attr:`SplitScreen.bottom_box`. 
    """
    def __init__(self, conf_settings=None, show_top_line_numbers=False,
            show_bottom_line_numbers=False):
        """Constructor

        :param conf_settings: a settings dictionary object. Defaults to 
                              `None` which uses the default settings
                              dictionary: :attr:`settings.settings`

        :param show_line_numbers: True turns line numbers on inside the
                                  associated code box. Defaults to False
        """
        self.show_top_line_numbers = show_top_line_numbers
        self.show_bottom_line_numbers = show_bottom_line_numbers
        super().__init__(conf_settings)

    def _build_boxes(self):
        # override the default build, creating two code boxes instead
        self.top_box = CodeBox(self, self.show_top_line_numbers)
        divider = (3, DividingLine())
        self.bottom_box = CodeBox(self, self.show_bottom_line_numbers)

        self.base_window = BaseWindow(self, [self.top_box, divider,
            self.bottom_box])

# =============================================================================
# Widgets
# =============================================================================

class AppendableText(urwid.AttrMap):
    ### Text-like widget that supports highlighting

    def __init__(self, markup):
        self.text_widget = urwid.Text('')
        self.set_text(markup)
        super(AppendableText, self).__init__(self.text_widget, 'empty')

    def set_text(self, markup):
        # urwid.Text supports three formats for markup: 1) plain text, 2)
        # tuple of palette attribute name and text, or 3) a list contain items
        # which are #1 or #2; to make things easier, convert it to a list
        self.markup = markup
        if isinstance(markup, str) or isinstance(markup, tuple):
            self.markup = [markup]

        self.text_widget.set_text(self.markup)

    def append(self, markup):
        self.markup.append(markup)
        self.set_text(self.markup)

    def set_highlight(self, highlight):
        if highlight:
            global highlight_mapper
            self.set_attr_map(highlight_mapper)
        else:
            self.set_attr_map({})

# -----------------------------------------------------------------------------

class DividingLine(urwid.Filler):
    tab_focusable = False

    def __init__(self):
        divider = urwid.Divider('-')
        super(DividingLine, self).__init__(divider, valign='top', top=1, 
            bottom=1)

# -----------------------------------------------------------------------------
# Code Box -- box that displays code

class ScrollingIndicator(urwid.Frame):
    def __init__(self):
        self.up = urwid.Text(' ')
        self.down = urwid.Text(' ')

        # create this Frame with a solid fill in the middle and the up and
        # down Text widgets as the header and footer
        super(ScrollingIndicator, self).__init__(urwid.SolidFill(' '), 
            header=self.up, footer=self.down)

    def set_up(self, is_up, focus):
        if is_up and focus:
            self.up.set_text('▲')
        elif is_up and not focus:
            self.up.set_text('△')
        else:
            self.up.set_text(' ')

    def set_down(self, is_down, focus):
        if is_down and focus:
            self.down.set_text('▼')
        elif is_down and not focus:
            self.down.set_text('▽')
        else:
            self.down.set_text(' ')


class ScrollingListBox(urwid.ListBox):
    def __init__(self, scroll_indicator, *args, **kwargs):
        self.scroll_indicator = scroll_indicator
        super(ScrollingListBox, self).__init__(*args, **kwargs)

    def render(self, size, focus):
        result = super(ScrollingListBox, self).render(size, focus)

        # ends_visible() returns a list with the words "top" and/or "bottom"
        # in it if the top and/or bottom of the list box is visible
        #
        # set our scrolling indicators based on what is visible
        visible = self.ends_visible(size)
        self.scroll_indicator.set_up('top' not in visible, focus)
        self.scroll_indicator.set_down('bottom' not in visible, focus)

        return result


class CodeBox(urwid.Columns):
    """Widget that displays the code. Actions use the methods in this class to
    change the text being displayed.

    The widget wraps an urwid ListBox, with each line in the box being a line
    of code. It also provides indiciators on the right side of the screen as
    to whether there is content above or below the current screen. If the
    parent :class:`Screen` implementation has multiple instances of this class
    active, the scroll area will also indicate which code box is focused. 

    The up and down arrows as well as the page-up and page-down buttons are
    supported. If there are multiple code boxes, tab key will change the
    focus.

    Each line of the code box displays text through the :class:`urwid.Text`
    widget. The methods in the this class for displaying content use urwid's
    markup tuple for determing the appearance of text. If you are writing new
    actions, you generally don't need to understand the markup language, just
    pass in the :class:`CodeLine.markup <purdy.content.CodeLine>` 
    attribute from the line in your :class:`Code <purdy.content.Code>` object.
    """
    tab_focusable = True

    # CodeBox is ListBox of Text with code in it accompanied by a side bar
    # with indicators about focus and scroll position
    def __init__(self, screen, show_line_numbers):
        """Constructor. These objects should only be constructed by a parent
        :class:`Screen` object.

        :param screen: the :class:`Screen` building this code box

        :param show_line_numbers: True if this box is to display line numbers
        """
        self.screen = screen
        self.show_line_numbers = show_line_numbers
        self.line_number = 1
        self.body = urwid.SimpleListWalker([])

        scroller = ScrollingIndicator()
        listbox = ScrollingListBox(scroller, self.body)

        layout = [listbox, (1, scroller)]

        super(CodeBox, self).__init__(layout)

    def clear(self):
        # clears the list and starts fresh
        self.body.contents.clear() 

    # --- Add Lines
    def add_line(self, markup=['']):
        """Appends an empty line to the end of the code listing

        :param markup: markup to put in the line, defaults to empty
        """
        self.add_line_at(len(self.body.contents) + 1, markup)

    def add_line_at(self, position, markup=['']):
        """Inserts a new empty line in the code box

        :param position: line number to insert at (1-indexed), pushes any
            existing content down; i.e. position=1 inserts at the top of the
            list
        :param markup: markup to put in the newly inserted line, defaults to
                       empty
        """
        if self.show_line_numbers:
            markup = [TokenLookup.line_number_markup(position), ] +  markup

        self.body.contents.insert(position - 1, AppendableText(markup))

    def fix_line_numbers(self, position):
        """Fixes line numbers after adding new lines to the code box"""
        line_no = position
        for widget in self.body.contents[line_no - 1:]:
            markup = widget.markup

            # this should only be called in show_line_numbers mode, so we can
            # assume that the first part of the markup is the line number
            # chunk
            #
            # re-build the markup, replacing the line number chunk with a new
            # one
            text = TokenLookup.line_number_markup(line_no)
            new_markup = [text, ] + markup[1:]
            widget.set_text(new_markup)

            line_no += 1

    # --- Set line
    def set_last_line(self, markup):
        """Changes the last line in the code box to have the given markup 

        :param markup: new markup tuple for the line
        """
        self.body.contents[-1].set_text(markup)

    def set_line(self, position, markup):
        """Changes the given line in the code box to have the given markup 

        :param position: line number to change (1-indexed)
        :param markup: new markup tuple for the line
        """
        self.body.contents[position - 1].set_text(markup)



import os, sys, re

from purdy import __version__

readme = os.path.join(os.path.dirname(__file__), 'README.rst')
long_description = open(readme).read()


SETUP_ARGS = dict(
    name='purdy',
    version=__version__,
    description=('Terminal based code snippet display tool '),
    long_description=long_description,
    url='https://github.com/cltrudeau/purdy',
    author='Christopher Trudeau',
    author_email='ctrudeau+pypi@arsensa.com',
    license='MIT',
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='code display',
    py_modules = ['purdy',],
    scripts=['purdy/purdy'],
    install_requires = [
        'Pygments>=2.4.2',
        'urwid>=2.0.1',
    ],
)

if __name__ == '__main__':
    from setuptools import setup, find_packages

    SETUP_ARGS['packages'] = find_packages()
    setup(**SETUP_ARGS)
